%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dafx2d.m
%
% Brian Hamilton, 2014
%
% This script should be called from dafxtests.m
% 
% This script performs a finite difference approximation of the 2-D wave equation
% with real-admittance boundaries (frequency-independent wall losses).  
% 
% The scheme stypes are standard finite difference (FD), Digital waveguide mesh
% (DWM), and finite volume (FV). The boundaries are approximated using 'centered'
% and 'non-centered' spatial differences as well as the '1-D' DWM boundaries.
% The re-entrant corner is treated in the standard fashion and in the case of %
% the 'centered' boundaries a three-quarter cell is also used, as given in the
% paper.
%
% The FV equivalent interpretation is also independently calculated and checked
% against the standard DWM and FD updates.  The energy is also calculated based
% on the FV geometries, which verifies the FD-FV equivalences established in
% paper for this particular scheme and domain.
% 
% Also note, this script is *not* optimised for speed!
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

try 
   clearvars -except stype gamma N domain; %MATLAB
catch 
   clear -x stype gamma N domain; %GNU Octave
end

switch domain
   case 'A'
      Ny    = 22;        % first dim, y                                               
      Nx    = 22;        % second dim, x                                              
      cory  =  8;        % re-entrant corner, y, domain A                             
      corx  =  8;        % re-entrant corner, x, domain A                             
   case 'B'
      Ny    = 22;        % first dim, y                                               
      Nx    = 22;        % second dim, x                                              
      cory  = 14;        % re-entrant corner, y, domain A                             
      corx  = 14;        % re-entrant corner, x, domain A                             
end

draw  = 0;         % set to 1 to plot simulation over time
c     = 1;         % wave speed (normalised)
l     = sqrt(0.5); % Courant number
l2    = 0.5;       % Courant number squared
outy  = Ny-3;      % output location, y
outx  = Nx-3;      % output location, x

h     = 1;         % spatial step
k     = l*h/c;     % time-step

%gamma prime, for DWM update
gammap = 2*gamma;

%linear index
coryx = cory+(corx-1)*Ny;

%states
u=zeros(Ny,Nx);
uz=zeros(Ny,Nx);
uzz=zeros(Ny,Nx);
%extra state for FV update (for verification)
ufv=zeros(Ny,Nx);

%initial conditions
uz(Ny-3,Nx-5)=1;
uz(Ny-3,Nx-3)=-1;

%initial energy
Hzero=2;

%total energy 
Hi=zeros(1,N);
%kinetic energy
TE=zeros(1,N);
%potential energy
VE=zeros(1,N);
%lost energy
Q=zeros(1,N);

if (draw)
   %plotting
   [X,Y]=meshgrid(1:Nx,1:Ny);
   h1=surf(X,Y,uz);
   axis equal
   shading interp
   view(3)
   set(gca,'DataAspectRatio',[1 1 0.15])
   %draw boundary
   ct=[[Ny-1,Nx-1];[2,Nx-1];[2,corx];[cory,corx];[cory,2];[Ny-1,2];[Ny-1,Nx-1]];
   hold on
   plot(ct(:,1),ct(:,2),'k-');
   hold off
end

%output signal
out=zeros(1,N);
%for difference between FD and FV calculations
e=zeros(1,N);

%mask
inside = zeros(Ny,Nx);
for y=2:Ny-1
   for x=2:Nx-1
      if x>=corx || y>=cory
         inside(y,x)=1;
      end
   end
end
outside = ~inside;

%for cell volumes
V=double(inside(:));
%for cell sides
S=ones(Nx*Ny,4);
%for cell boundary sides
Sb=zeros(Nx*Ny,1);
%neighbour mask
Nm=repmat(double(inside(:)),1,4);

%for interior points with Sb=0
Ii=zeros(size(outside));
%for stencil weights
sw=zeros(length(u(:)),4); 

switch stype
   case {'FDnc','DWM'}
      ew=1; %edge weight
   case {'FDc','FDc3q'}
      ew=2; %edge weight
end

%now update V,S,Sb,Nm,Ii,sw
for y=2:Ny-1
   for x=2:Nx-1
      if (inside(y,x))
         i = y+(x-1)*Ny;
         sw(i,1:4)=1;
         if (inside(y-1,x) && inside(y+1,x) && inside(y,x-1) && inside(y,x+1)) 
            Ii(y,x)=1;
         end
         if (outside(y-1,x)) 
            sw(i,2)=ew;
            sw(i,1)=0;
            Nm(i,1)=0;
            switch stype; case {'FDc3q'}
               S(i,3:4)=0.5;
               V(i)=V(i)/2;
               Sb(i)=1;
            case 'FDnc'
               Sb(i)=Sb(i)+1;
            case 'DWM'
               V(i)=1/4;
               Nm(i,[1,3,4])=0;
               Sb(i)=sqrt(2);
            end
         end
         if (outside(y+1,x))
            sw(i,1)=ew;
            sw(i,2)=0;
            Nm(i,2)=0;
            switch stype; case {'FDc3q'}
               S(i,3:4)=0.5;
               V(i)=V(i)/2;
               Sb(i)=1;
            case 'FDnc'
               Sb(i)=Sb(i)+1;
            case 'DWM'
               V(i)=1/4;
               Nm(i,[2,3,4])=0;
               Sb(i)=sqrt(2);
            end
         end
         if (outside(y,x-1))
            sw(i,4)=ew;
            sw(i,3)=0;
            Nm(i,3)=0;
            switch stype; case {'FDc3q'}
               S(i,1:2)=0.5;
               V(i)=V(i)/2;
               Sb(i)=1;
            case 'FDnc'
               Sb(i)=Sb(i)+1;
            case 'DWM'
               V(i)=1/4;
               Nm(i,[1,2,3])=0;
               Sb(i)=sqrt(2);
            end
         end
         if (outside(y,x+1))
            sw(i,3)=ew;
            sw(i,4)=0;
            Nm(i,4)=0;
            switch stype; case {'FDc3q'}
               S(i,1:2)=0.5;
               V(i)=V(i)/2;
               Sb(i)=1;
            case 'FDnc'
               Sb(i)=Sb(i)+1;
            case 'DWM'
               V(i)=1/4;
               Nm(i,[1,2,4])=0;
               Sb(i)=sqrt(2);
            end
         end
      end
   end
end

%fix re-entrant corner
switch stype
   case 'FDc3q'
      V(coryx)=3/4;
      S(coryx,1)=0.5;
      S(coryx,2)=1;
      S(coryx,3)=0.5;
      S(coryx,4)=1;
      Sb(coryx)=1;
      Ii(coryx)=0;
   case 'DWM'
      V(coryx)=1/2;
      S(coryx,1)=0;
      S(coryx,2)=1;
      S(coryx,3)=0;
      S(coryx,4)=1;
      Sb(coryx)=sqrt(2);
      Ii(coryx)=0;
end

%scale by h and mask
S=S*h.*Nm;
%scale by h
Sb=Sb*h;
%scale by h^2
V=V*h^2;


for n=1:N %time-loop
   for x=2:Nx-1 %x-loop
      for y=2:Ny-1 %y-loop
         i = y+(x-1)*Ny;
         if (inside(i)) %if interior or boundary cell
            if (Ii(y,x)) %regular update (FD/DWM)
               u(y,x) = l2*(uz(y-1,x) +uz(y+1,x) +uz(y,x-1) +uz(y,x+1)) - uzz(y,x);
            else %boundary nodes
               switch stype
                  case 'DWM'
                     K=Ii(y+1,x)+Ii(y-1,x)+Ii(y,x+1)+Ii(y,x-1); %number of connected air nodes
                     if (K>0) 
                        u(y,x) = (1/(K+gammap))*(2*(Ii(y-1,x)*uz(y-1,x)+Ii(y+1,x)*uz(y+1,x)+Ii(y,x-1)*uz(y,x-1)+Ii(y,x+1)*uz(y,x+1)) + (gammap-K)*uzz(y,x));
                     end
                  case 'FDnc'
                     Ng=sum(sw(i,:)==0); %number of ghost points
                     u(y,x) = (1/(1+l*gamma/2*Ng))*(...
                        l2*(sum(sw(i,:).*[uz(y-1,x),uz(y+1,x),uz(y,x-1),uz(y,x+1)])-sum(sw(i,:))*uz(y,x)) + 2*uz(y,x) + (l*gamma/2*Ng-1)*uzz(y,x));
                  case {'FDc','FDc3q'}
                     Ng=sum(sw(i,:)==0); %number of ghost points
                     u(y,x) = (1/(1+l*gamma*Ng))*(...
                        l2*(sum(sw(i,:).*[uz(y-1,x),uz(y+1,x),uz(y,x-1),uz(y,x+1)])-sum(sw(i,:))*uz(y,x)) + 2*uz(y,x) + (l*gamma*Ng-1)*uzz(y,x));
               end
            end
            %three-quarter cell update
            if (y==cory && x==corx) && strcmp(stype,'FDc3q')
               u(y,x) = (1/(1+l*gamma*2/3))*(...
                  l2*(sum([2/3 4/3 2/3 4/3].*[uz(y-1,x),uz(y+1,x),uz(y,x-1),uz(y,x+1)])-4*uz(y,x)) + 2*uz(y,x) + (l*gamma*2/3-1)*uzz(y,x));
            end

            if ~strcmp(stype,'FDc')
               %recalculate using FV update and equivalent interpretation
               %using FV cell parameters
               sigma=c*k*Sb(i)/2/V(i);
               isg1=1/(1+sigma*gamma);
               sg2=(sigma*gamma-1);
               ufv(i) = isg1*(l2*h/V(i)*(...
                  S(i,1)*(uz(i-1)  - uz(i))...
                  +S(i,2)*(uz(i+1)  - uz(i))...
                  +S(i,3)*(uz(i-Ny) - uz(i))...
                  +S(i,4)*(uz(i+Ny) - uz(i)))...
                  +2*uz(i) + sg2*uzz(i));

               %calculate energy
               TE(n)=TE(n)+V(i)/2/c^2/k^2*(uz(i)-uzz(i))*(uz(i)-uzz(i));
               VE(n)=VE(n)+1/4/h*(...
                   S(i,1)*(uz(i)-uz(i-1) )*(uzz(i)-uzz(i-1) )...
                  +S(i,2)*(uz(i)-uz(i+1) )*(uzz(i)-uzz(i+1) )...
                  +S(i,3)*(uz(i)-uz(i-Ny))*(uzz(i)-uzz(i-Ny))...
                  +S(i,4)*(uz(i)-uz(i+Ny))*(uzz(i)-uzz(i+Ny)));
               Q(n)=Q(n)+1/c*gamma*Sb(i)*1/4/k^2*(u(i)-uzz(i))^2;
            end
         end
      end
   end
   %total energy
   Hi(n)=TE(n)+VE(n);
   if (draw)
      set(h1,'zdata',u);drawnow;
   end
   if ~strcmp(stype,'FDc')
      e(n)=norm(ufv-u);
   end
   %output
   out(n)=u(outy,outx);

   %time-stepping
   uzz=uz;
   uz=u;
end


if strcmp(stype,'FDc')
   %plot output from one node
   figure
   plot(out);
   xlabel('time-steps')
   title(['output, domain ',domain,', ',stype])
%plot energy, FDc left out since energy expression not available
else
   %plot output from one node
   figure
   subplot(2,2,1)
   plot(out);
   xlabel('time-steps')
   title(['output, domain ',domain,', ',stype])

   %plot difference between FD and FV equivalent updates
   subplot(2,2,3)
   plot(e,'k.');
   xlabel('time-steps')
   title('difference between FD-based calculations and FV equivalent calculations')

   subplot(2,2,2)
   plot(Hi,'k.');
   hold on
   plot(TE,'g.');
   plot(VE,'b.');
   hold off
   if gamma>0
      hold on
      plot(Q,'r.');
      plot(1/k*diff(Hi),'c.');
      hold off
      legend({'Hi=TE+VE','TE','VE','wall losses','d/dt Hi'},'location','best');
   else
      legend({'Hi=TE+VE','TE','VE'},'location','best');
   end
   xlabel('time-steps')
   title(['energy, domain ',domain,', ',stype])

   subplot(2,2,4)
   if gamma==0
      %lossless, plot energy variation
      plot((Hi-Hzero)./Hzero,'k.');
      title('(Hi(n)-Hi(0))./Hi(0)')
      xlabel('time-steps')
   else
      %lossy, plot rate of change of energy and losses
      plot((1/k*diff(Hi)+Q(1:end-1)),'k.');
      title('wall losses + change in energy')
      xlabel('time-steps')
   end
end
